#!mkcmd
# $Mkcmd(*): ${mkcmd-mkcmd} %f
# mkcmd script for explode's user interface
#
from '<stdlib.h>'
from '"machine.h"'
from '"explode.h"'
from '<sys/param.h>'
from '<sys/file.h>'
from '<sysexits.h>'
from '<signal.h>'
from '<ctype.h>'
from '<stdio.h>'
from '<errno.h>'
from '<pwd.h>'

require "std_help.m" "std_version.m" "api_explmux.m"

%i
static const char
	rcsid[] = "$Id: explode.m,v 6.10 2013/06/06 17:58:50 ksb Exp $";
extern void Version(FILE *);
%%
%h
extern void SetDelExt(/* char * */);
%%
string named "acDefIncl" {
	init 'DEFINCL'
}

getenv "EXPLODE"
basename "explode" ""

boolean 't' {
	exclude "nfisTH"
	named "fTable"
	help "provide a table of contents output (like tar)"
}
boolean 's' {
	exclude "cduTH"
	named "fWhole"
	help "output the source file rather than the parts"
	boolean 'S' {
		named "fPrint"
		help "just display the source path"
	}
}
boolean 'c' {
	exclude "st"
	named "fHeader"
	init "1"
	help "do not include any Header section in the output"
}

boolean 'f' {
	named "fOverwrite"
	initializer "0"
	help "force overwrite of files that exist"
}
boolean 'n' {
	named "fActive"
	initializer "1"
	help "report what would be done, but don't do it"
}
boolean 'v' {
	named "fVerbose"
	help "report actions as they happen"
}
char* 'e' {
	named "pcExt"
	parameter "extender"
	help "specify an extender for all created files"
}
char* 'd' {
	named "pcDel"
	parameter "delimiter"
	help "specify a different delimiter (rather than /*@)"
}
boolean 'i' {
	named "fAsk"
	help "be interactive, like rm"
}
char* 'o' {
	named "pcName"
	parameter "prefix"
	help "specify a prefix for all created filenames"
}
accum[":"] 'I' {
	named "pcIncl"
	param "dir"
	help "a colon separated list of search directories for source files"
}
accum[","] 'u' {
	named "pcUnpack"
	parameter "unpack"
	help "a comma separated list of components to unpack"
}
char* 'R' {
	named "pcRemote"
	param "remote"
	help "consult a network peer's tcpmux for modules, optional :explmux allowed"
	exclude "THV"
}

augment action 'V' {
	user 'printf("%%s: default repository %%s\\n", %b, acDefIncl);VersionExplmux(stdout);'
}
action 'T' {
	named "DumpTable"
	update "%n(stdout);"
	help "list the recognized file extenders"
	abort "exit(0);"
}
action 'H' {
	named "Details"
	update "%n(stdout);"
	help "give some details of the file format expected"
	abort "exit(0);"
}

after {
	named "SetGlobals"
	update '%n();%rIn = optaccum(%rIn, acDefIncl, "%rId");'
}
every {
	named "UnPack"
	update "%n(%a, %ron, %rRn);"
	parameter "files"
	help "the files to unpack"
}
zero {
	update "%Ren((char *)0, %ron, (char *)0);"
}
exit {
	named "RemoveTemp"
	update "%n();"
}

int variable "iDelLen" {
	init "3"
	help "escape delimiter size"
}
int variable "cClose" {
	init "'@'"
	help "close an Explode directive"
}

# implementation details
%i
typedef struct LInode {
	char *pclang;		/* which language			*/
	char *pccomment;	/* the comment introducer		*/
	char *pcext;		/* the extender we expect		*/
	char cclose;		/* if not last character in comment	*/
} LANG_INFO;
%%

%c
static LANG_INFO LITable[] = {
	{"C",		"/*@",	".c",	'@'},
	{"Bourne Shell","#@",	".sh",	'@'},
	{"C Shell",	"#@",	".csh",	'@'},
	{"C header",	"/*@",	".h",	'@'},
	{"C++ implementation",	"//@",	".ccP",	'@'},
	{"C++ interface",	"//@",	".hP",	'@'},
	{"FORTRAN",	"c@",	".f",	'@'},
	{"include",	"/*@",	".i",	'@'},
	{"K",		"/*@",	".k",	'@'},
	{"Pascal",	"(*@",	".p",	'@'},	/* could  be {@ too @}) */
	{"Tcl",		"; @",	".tcl",	'@'},
	{"elisp",	";@",	".el",	'@'},	/* comment correct? */
	{"lex",		"/*@",	".l",	'@'},	/* also work for bc and dc */
	{"html",	"<!--@",".htm", '@'},
	{"html",	"<!--@",".html",'@'},
	{"hxmd",	"#@",   ".hxmd",'@'},
	{"m4",		"dnl@", ".host",'@'},
	{"m4",		"dnl@", ".m4",	'@'},
	{"mkcmd",	"#@",	".m",	'@'},
	{"mkcmd",	"/*@",	".mc",	'@'},
	{"mkcmd",	"/*@",	".mh",	'@'},
	{"mkcmd",	"/*@",	".mi",	'@'},
	{"make",	"#@",	".mk",	'@'},
	{"perl",	"#@",	".pl",	'@'},
	{"ray shade",	"/*@",	".ray",	'@'},
	{"tcl",		"#@",	".tcl",	'@'},
	{"xml",		"<!--@",".xml", '@'},
	{"X tool kit",	"#@",	".tk",	'@'},
	{"X tool kit",	"#@",	".wish",'@'},
	{"yacc",	"/*@",	".y",	'@'},
	{"nroff",	".\\\"@", ".man", '@'},
	{"nroff",	".\\\"@", ".ms",  '@'},
	{"nroff",	".\\\"@", ".me",  '@'},
	{"nroff",	".\\\"@", ".mm",  '@'},
	{"nroff",	".\\\"@", ".nro", '@'},
};
#define MAXLANG		(sizeof(LITable)/sizeof(struct LInode))

/* output the table above to the user					(ksb)
 */
static void
DumpTable(FILE *fp)
{
	register int i;
	register unsigned u1, u2;

	u1 = u2 = 7;
	for (i = 0; i < MAXLANG; ++i) {
		register unsigned u;
		u = strlen(LITable[i].pclang);
		if (u > u1) {
			u1 = u;
		}
		u = strlen(LITable[i].pccomment);
		if (u > u2) {
			u2 = u;
		}
	}
	for (i = 0; i < MAXLANG; ++i) {
		fprintf(fp, "%-*s %-*s %s\n", u1, LITable[i].pclang, u2, LITable[i].pccomment, LITable[i].pcext);
	}
}

static char *apchHelp[] = {
	"%sAppend file    divert text to the end of file\n",
	"%sHeader         divert text to the header file\n",
	"%sExplode file   copy header to file then divert text to file\n",
	"%sRemove         divert text to /dev/null\n",
	"%sInsert string  insert text in output stream\n",
	"%sShell string   sh(1) text, put output in stream\n",
	"%sVersion string mark the version of a module\n",
	"%sMessage        divert text to /dev/tty\n",
	(char *)0
};

/* a better help for the user						(ksb)
 * This should be enough
 */
static void
Details(FILE *fp)
{
	register char **ppc, *pcLine, *pcTemp;

	pcTemp = (char*)0 != pcDel ? pcDel : LITable[0].pccomment;
	for (ppc = & apchHelp[0]; (char *)0 != (pcLine = *ppc); ++ppc) {
		fprintf(fp, pcLine, pcTemp);
	}
}

/* Set the delimiter from the file extender				(ksb)
 */
void
SetDelExt(char *pcFile)
{
	static int fFakeExt = 0;
	static int fFakeDel = 0;
	register char *pc, *pcHas;
	register int i = MAXLANG;

	pcHas = strrchr(pcFile, '.');
	if (((char *)0 != (pc = pcExt) && !fFakeExt) || (char *)0 != (pc = pcHas)) {
		for (i = 0; i < MAXLANG; ++i) {
			if (0 == strcmp(LITable[i].pcext, pc)) {
				break;
			}
		}
	}

	if (MAXLANG == i) {
		i = 0;			/* default to C		*/
		LITable[0].pcext = (char *)0 != pcHas ? pcHas : ".c";
	}

	if ((char *)0 == pcExt || fFakeExt) {
		fFakeExt = 1;
		pcExt = LITable[i].pcext;
	}

	if ((char *)0 == pcDel || fFakeDel) {
		fFakeDel = 1;
		pcDel = LITable[i].pccomment;
	}
	iDelLen = strlen(pcDel);
	if ('\000' == (cClose = LITable[i].cclose)) {
		cClose = pcDel[iDelLen - 1];
	}
}


#include "machine.h"
#include "explode.h"
#include "main.h"

extern struct passwd *getpwnam();

#if HAVE_STRINGS
#include <strings.h>
#else
#include <string.h>
#endif

/* we have to die because we are out of core				(ksb)
 */
static void
OutOfMem()
{
	static char acFailMem[] = ": out of memory\n";

	write(2, progname, strlen(progname));
	write(2, acFailMem, sizeof(acFailMem)-1);
	exit(ENOMEM);
}

static const char
	acNull[]= "/dev/null",	/* the bit bucket			*/
	acTty[]= "/dev/tty",	/* the users tty			*/
	acAppend[]= "Append",
	acHead[]= "Header",
	acExp[] = "Explode",
	acRem[] = "Remove",
	acVersion[] = "Version",
	acIns[] = "Insert",
	acShell[] = "Shell",
	acMsg[] = "Message";
#define MAXLINE		2048	/* max C code line ANSI sez 509		*/

static char
	*pcTemp,		/* the real temp file name		*/
	acTemp[] =		/* file name template for header	*/
		"/tmp/explXXXXXX",
	acStdin[] =		/* internal name for "-" on cmd line	*/
		"stdin";

/* search a path (colon separated) for a directory that has the file	(ksb)
 * fall back on ./file always first, and full paths are not scanned
 */
static FILE *
SearchOpen(char *pcPath, char *pcFile, char *pcHow)
{
	register char *pcNext;
	static char acColon[] = ":";
	register FILE *fpFound;
	auto char acTry[MAXPATHLEN+1];

	if (acStdin == pcFile) {
		if (fPrint) {
			printf("-\n");
		}
		return stdin;
	}
	if ((FILE *)0 != (fpFound = fopen(pcFile, pcHow)) || '/' == pcFile[0]) {
		if (fPrint && (FILE *)0 != fpFound) {
			printf("%s\n", pcFile);
		}
		return fpFound;
	}

	for (; *pcPath != '\000'; (*pcNext = ':'), pcPath = pcNext+1) {
		pcNext = strchr(pcPath, ':');
		if ((char *)0 == pcNext)
			pcNext = acColon;
		*pcNext = '\000';

		if ('~' == pcPath[0]) {
			register char *pcSlash;
			register struct passwd *pwd;
			static char acSlash[] = "/";

			if ((char *)0 == (pcSlash = strchr(pcPath+1, '/')))
				pcSlash = acSlash;
			*pcSlash = '\000';
			pwd = ('\000' == pcPath[1]) ? getpwuid(getuid()) : getpwnam(pcPath+1);
			*pcSlash = '/';
			if ((struct passwd *)0 == pwd) {
				continue;
			}
			sprintf(acTry, "%s/%s/%s", pwd->pw_dir, pcSlash+1, pcFile);
		} else {
			sprintf(acTry, "%s/%s", pcPath, pcFile);
		}
		if (fVerbose) {
			fprintf(stderr, "%s: search for %s as %s\n", progname, pcFile, acTry);
		}
		if ((FILE *)0 != (fpFound = fopen(acTry, pcHow))) {
			*pcNext = ':';
			if (fPrint) {
				printf("%s\n", acTry);
			}
			return fpFound;
		}
	}
	return (FILE *)0;
}

/* make sure global flages are consistant				(ksb)
 * and build temp file name
 */
void
SetGlobals()
{
	extern char *mktemp();
	register int iFd;

	if (!fActive) {
		fVerbose = 1;
	}
#if USE_MKSTEMP
	if (-1 == (iFd = mkstemp(acTemp))) {
		fprintf(stderr, "%s: mkstemp: %s\n", progname, strerror(errno));
		exit(EX_TEMPFAIL);
	}
	pcTemp = acTemp;
#else
	/* Try to make mktemp as safe as we can				(ksb)
	 */
	if ((char *)0 == (pcTemp = mktemp(acTemp))) {
		fprintf(stderr, "%s: mktemp: %s\n", progname, strerror(errno));
		exit(EX_TEMPFAIL);
	}
	if (-1 == (iFd = open(pcTemp, O_RDWR, 0600))) {
		fprintf(stderr, "%s: open: %s: %s\n", progname, pcTemp, strerror(errno));
		(void)unlink(pcTemp);
		exit(EX_TEMPFAIL);
	}
	if (-1 == fchmod(iFd, 0600)) {
		fprintf(stderr, "%s: fchmod: %s: %s\n", progname, pcTemp, strerror(errno));
		(void)unlink(pcTemp);
		exit(EX_TEMPFAIL);
	}
#endif
	(void)close(iFd);
}


/* zap the temp file							(ksb)
 */
void
RemoveTemp()
{
	(void)unlink(pcTemp);
}

/* tell if we need this file						(ksb)
 * We need all files if none are asked for, we need only those
 * in the list is any are asked for.
 */
static int
Need(char *pc)
{
	register char *pcCur, *pcNext;
	static char acComma[] = ",";

	if ((char *)0 == (pcCur = pcUnpack)) {
		return 1;
	}
	for (; *pcCur != '\000'; (*pcNext = ','), pcCur = pcNext+1) {
		pcNext = strchr(pcCur, ',');
		if ((char *)0 == pcNext)
			pcNext = acComma;
		*pcNext = '\000';

		if (0 == strcmp(pc, pcCur)) {
			*pcNext = ',';
			return 1;
		}
	}
	return 0;
}

/* qwery the user WRT this file, should we alter it			(ksb)
 */
static int
Asked(char *pcFile)
{
	static char *pcTold = (char *)0;
	auto char acAns[32];
	register char *pcCur, *pcNext;
	static char acComma[] = ",";

	if (!fAsk) {
		return 0;
	}
	/* is this in the list we already asked about?
	 */
	if ((char *)0 != (pcCur = pcTold)) {
		for (pcCur = pcTold; *pcCur != '\000'; (*pcNext = ','), pcCur = pcNext+1) {
			pcNext = strchr(pcCur, ',');
			if ((char *)0 == pcNext)
				pcNext = acComma;
			*pcNext = '\000';
			if (0 == strcmp(pcFile, pcCur+1)) {
				*pcNext = ',';
				return 'y' == *pcCur;
			}
		}
	}
	fprintf(stdout, "%s: modify %s? [ny] ", progname, pcFile);
	fflush(stdout);
	if ((char *)0 == fgets(acAns, sizeof(acAns), stdin)) {
		fprintf(stdout, "no\n");
		acAns[0] = 'n';
	}
	for (pcCur = acAns; isspace(*pcCur); ++pcCur)
		/* nothing */;

	acAns[0] = ('Y' == *pcCur || 'y' == *pcCur) ? 'y' : 'n';

	if ((char *)0 == pcTold) {
		pcTold = pcCur = malloc(1+strlen(pcFile)+1);
		if ((char *)0 == pcTold) {
			OutOfMem();
		}
	} else {
		register int i;
		i = strlen(pcTold);
		pcTold = realloc(pcTold, i+1+1+strlen(pcFile)+1);
		if ((char *)0 == pcTold) {
			OutOfMem();
		}
		pcCur = pcTold + i;
		*pcCur++ = ',';
	}
	*pcCur++ = acAns[0];
	(void)strcpy(pcCur, pcFile);
	return 'y' == acAns[0];
}

/* does all the work of unpacking the files from the MICE archive	(ksb)
 */
void
UnPack(char *pcFile, char *pcBase, const char *pcRemote)
{
	static char acPath[MAXPATHLEN+1], acWMode[] = "w", acAMode[] = "a";
	auto char acLine[MAXLINE+1];
	auto char acBase[255];
	register FILE *fpIn, *fpOut;
	register char *pcMode, *pc;
	register FILE *fpHead;
	register int fNewHeader;

	if ((char *)0 == pcFile || ('-' == pcFile[0] && '\000' == pcFile[1])) {
		pcFile = acStdin;
		if ((char *)0 == pcBase) {
			pcBase = acStdin;
		}
	} else if ((char *)0 == pcBase) {
		pc = strrchr(pcFile, '/');
		if ((char *)0 == pc)
			pc = pcFile;
		else
			++pc;
		(void)strcpy(acBase, pc);
		pc = strrchr(acBase, '.');
		if ((char *)0 != pc)
			*pc = '\000';
		pcBase = acBase;
	}

	SetDelExt(pcFile);

	if (fVerbose) {
		if (fTable)
			fprintf(stdout, "%s: contents of ", progname);
		else
			fprintf(stdout, "%s: unpack ", progname);
		if ((char *)0 != pcUnpack)
			fprintf(stdout, "only %s in ", pcUnpack);
		fprintf(stdout, "%s to %s*%s\n", acStdin == pcFile ? "<stdin>" : pcFile, pcBase, pcExt);
	}

	if ((const char *)0 != pcRemote) {
		if ((FILE *)0 == (fpIn = RemoteOpen(& pcRemote, pcFile, "rb"))) {
			fprintf(stderr, "%s: %s: %s: %s\n", progname, pcRemote, pcFile, strerror(ENOENT));
			exit(EX_NOINPUT);
		}
	}
	if ((FILE *)0 == (fpIn = SearchOpen(pcIncl, pcFile, "r"))) {
		fprintf(stderr, "%s: fopen: %s: %s\n", progname, pcFile, strerror(errno));
		exit(EX_NOINPUT);
	}
	if (fPrint) {
		return;
	}

	fpOut = (FILE *)0;
	if (fWhole) {
		if ('-' != pcBase[0] || '\000' != pcBase[1]) {
			sprintf(acPath, "%s%s", pcBase, pcExt);
			if (0 == fOverwrite && 0 == access(acPath, F_OK) && 0 == Asked(acPath)) {
				fprintf(stderr, "%s: won't overwrite %s\n", progname, acPath);
				exit(EX_DATAERR);
			}
			if (fVerbose) {
				fprintf(stdout, "%s: found %s\n", progname, acPath);
			}
			if (! fActive) {
				goto get_out;
			}
			if ((FILE *)0 == (fpOut = fopen(acPath, acWMode))) {
				fprintf(stderr, "%s: fopen: %s: %s\n", progname, acPath, strerror(errno));
				exit(EX_CANTCREAT);
			}
		} else if (! fActive) {
			goto get_out;
		} else {
			fpOut = stdout;
		}
		while ((char *)0 != fgets(acLine, sizeof(acLine), fpIn)) {
			fputs(acLine, fpOut);
		}
		goto get_out;
	}

	if ((FILE *)0 == (fpHead = fopen(pcTemp, "w"))) {
		fprintf(stderr, "%s: fopen: %s: %s\n", progname, pcTemp, strerror(errno));
		exit(EX_OSERR);
	}
#ifdef DEBUG
	fprintf(fpHead, "%s no header given */\n", pcDel);
#endif /* DEBUG */
	fclose(fpHead);

	if ((FILE *)0 == (fpOut = fopen(acNull, "w"))) {
		fprintf(stderr, "%s: fopen: %s: %s\n", progname, acNull, strerror(errno));
		exit(EX_OSERR);
	}
	fNewHeader = 0;
	while ((char *)0 != fgets(acLine, sizeof(acLine), fpIn)) {
		if (0 != strncmp(pcDel, acLine, iDelLen)) {
			fputs(acLine, fpOut);
			continue;
		}
		pc = & acLine[iDelLen];
		while (isspace(*pc))
			++pc;
		pcMode = acWMode;
		if (fTable) {
			register char *pcFind;
			static char acCurVer[2048];

			if (0 == strncmp(pc, acVersion, sizeof(acVersion)-1)) {
				pc += sizeof(acVersion)-1;
				while (isspace(*pc))
					++pc;
				pcFind = acCurVer;
				while (*pc != cClose)
					*pcFind++ = *pc++;
				*pcFind = '\000';
				continue;
			} else if (0 == strncmp(pc, acExp, sizeof(acExp)-1)) {
				pc += sizeof(acExp);
			} else if (0 == strncmp(pc, acHead, sizeof(acHead)-1)) {
				pc += sizeof(acHead);
			} else {
				continue;
			}
			while (isspace(*pc))
				++pc;
			pcFind = pc;
			while (!isspace(*pc) && *pc != cClose)
				++pc;
			*pc = '\000';
			if ('\000' == *pcFind)
				pcFind = "*";
			else if ((char *)0 != pcUnpack && ! Need(pcFind))
				continue;
			printf("%s\t%s%s%s", pcFile, pcBase, pcFind, pcExt);
			if ('\000' != acCurVer[0])
				printf("\t%s", acCurVer);
			printf("\n");
			continue;
		}
		if (0 == strncmp(pc, acExp, sizeof(acExp)-1) ||
		   (pcMode = acAMode,
		   0 == strncmp(pc, acAppend, sizeof(acAppend)-1))) {
			register char *pcFind;

			pc += pcMode == acExp ? sizeof(acExp) : sizeof(acAppend);
			while (isspace(*pc))
				++pc;
			pcFind = pc;
			while (!isspace(*pc) && *pc != cClose)
				++pc;
			*pc = '\000';
			if ((char *)0 != pcUnpack && ! Need(pcFind)) {
				goto dev_null;
			}
			if (stdout != fpOut) {
				fclose(fpOut);
			}
			if ('-' != pcBase[0] || '\000' != pcBase[1]) {
				sprintf(acPath, "%s%s%s", pcBase, pcFind, pcExt);
				if (0 == fOverwrite && acWMode == pcMode &&
				    0 == access(acPath, F_OK) &&
				    0 == Asked(acPath)) {
					fprintf(stderr, "%s: won't overwrite %s\n", progname, acPath);
					exit(EX_DATAERR);
				}
				if (fVerbose) {
					fprintf(stdout, "%s: %s %s\n", progname, acWMode == pcMode ? "create" : "append", acPath);
				}
				if (! fActive) {
					continue;
				}
				if ((FILE *)0 == (fpOut = fopen(acPath, pcMode))) {
					fprintf(stderr, "%s: fopen: %s: %s\n", progname, acPath, strerror(errno));
					exit(EX_CANTCREAT);
				}
			} else if (! fActive) {
				continue;
			} else {
				fpOut = stdout;
				if (! fNewHeader) {
					continue;
				}
			}
			if (acWMode != pcMode) {
				continue;
			}
			if ((FILE *)0 == (fpHead = fopen(pcTemp, "r"))) {
				fprintf(stderr, "%s: fopen: %s: %s\n", progname, pcTemp, strerror(errno));
				exit(EX_OSERR);
			}
			while ((char *)0 != fgets(acLine, sizeof(acLine), fpHead)) {
				fputs(acLine, fpOut);
			}
			fclose(fpHead);
			fNewHeader = 0;
		} else if (0 == strncmp(pc, acIns, sizeof(acIns)-1)) {
			register int iStop;

			if (fVerbose)
				fprintf(stdout, "%s: insert text\n", progname);
			pc += sizeof(acIns)-1;
			while (isspace(*pc))
				++pc;
			iStop = *pc;
			while (iStop != *++pc)
				putc(*pc, fpOut);
			putc('\n', fpOut);
		} else if (0 == strncmp(pc, acShell, sizeof(acShell)-1)) {
			register int iStop;
			register char *pcCmd;
			register FILE *fpShell;

			pc += sizeof(acShell)-1;
			while (isspace(*pc))
				++pc;
			iStop = *pc;
			pcCmd = pc+1;
			while (iStop != *++pc)
				;
			*pc = '\000';
			if (fVerbose)
				fprintf(stdout, "%s: shell `%s'\n", progname, pcCmd);
			if ((FILE *)0 == (fpShell = popen(pcCmd, "r"))) {
				fprintf(stdout, "%s: shell fail: %s\n", progname, strerror(errno));
				exit(EX_UNAVAILABLE);
			}
			while (EOF != (iStop = getc(fpShell))) {
				putc(iStop, fpOut);
			}
			pclose(fpShell);
		} else if (0 == strncmp(pc, acHead, sizeof(acHead)-1)) {
			if (fVerbose)
				fprintf(stdout, "%s: new header\n", progname);
			fNewHeader = fHeader;
			fflush(fpOut);
			if (stdout != fpOut)
				fclose(fpOut);
			if (fActive && (FILE *)0 == (fpOut = fopen(pcTemp, "w"))) {
				fprintf(stderr, "%s: fopen: %s: %s\n", progname, pcTemp, strerror(errno));
				exit(EX_OSERR);
			}
		} else if (0 == strncmp(pc, acRem, sizeof(acRem)-1)) {
			if (fVerbose)
				fprintf(stdout, "%s: remove text\n", progname);
			fflush(fpOut);
dev_null:
			if (stdout != fpOut)
				fclose(fpOut);
			if (fActive && (FILE *)0 == (fpOut = fopen(acNull, "w"))) {
				fprintf(stderr, "%s: fopen: %s: %s\n", progname, acNull, strerror(errno));
				exit(EX_OSERR);
			}
		} else if (0 == strncmp(pc, acMsg, sizeof(acMsg)-1)) {
			if (fVerbose)
				fprintf(stdout, "%s: message text\n", progname);
			fflush(fpOut);
			if (stdout != fpOut)
				fclose(fpOut);
			if (fActive && (FILE *)0 == (fpOut = fopen(acTty, "w"))) {
				fprintf(stderr, "%s: fopen: %s: %s\n", progname, acTty, strerror(errno));
				exit(EX_IOERR);
			}
		}
	}

get_out:
	if ((FILE *)0 != fpIn && stdin != fpIn) {
		fclose(fpIn);
	}
	if ((FILE *)0 != fpOut && stdout != fpOut)
		fclose(fpOut);
	RemoveTemp();
}
