.\" $Id: explode.man,v 6.25 2013/07/25 15:08:25 ksb Exp $
.\" $Compile: Display%h
.\" $Display: ${groff:-groff} -t -Tascii -man %f | ${PAGER:-less}
.\" $Display(*): ${groff:-groff} -t -T%s -man %f
.\" $Install: %b -mDeinstall %o %f && cp %f $DESTDIR${TOP:-/usr/local}/man/man1/explode.1
.\" $Deinstall: ${rm-rm} -f $DESTDIR${TOP:-/usr/local}/man/[cm]a[nt]1/explode.1*
.TH EXPLODE 1 LOCAL
.SH NAME
explode - unpack components from aggregated source archives
.SH SYNOPSIS
.ds PN "explode
\fI\*(PN\fP [\fB\-cfinv\fP] [\fB\-d\fP\~\fIdelimiter\fP] [\fB\-e\fP\~\fIextender\fP] [\fB\-I\fP\~\fIdir\fP] [\fB\-o\fP\~\fIprefix\fP] [\fB\-R\fP\~\fIremote\fP] [\fB\-u\fP\~\fIunpack\fP] [\fIfiles\fP]
.br
\fI\*(PN\fP [\fB\-tv\fP] [\fB\-d\fP\~\fIdelimiter\fP] [\fB\-e\fP\~\fIextender\fP] [\fB\-I\fP\~\fIdir\fP] [\fB\-o\fP\~\fIprefix\fP] [\fB\-R\fP\~\fIremote\fP] [\fB\-u\fP\~\fIunpack\fP] [\fIfiles\fP]
.br
\fI\*(PN\fP [\fB\-finv\fP] [\fB\-e\fP\~\fIextender\fP] [\fB\-I\fP\~\fIdir\fP] [\fB\-o\fP\~\fIprefix\fP] [\fB\-R\fP\~\fIremote\fP] [\fIfiles\fP]
.br
\fI\*(PN\fP \fB\-h\fP
.br
\fI\*(PN\fP \fB\-H\fP
.br
\fI\*(PN\fP \fB\-T\fP
.br
\fI\*(PN\fP \fB\-V\fP
.br
\fI\*(PN\fP \fB\-s\fP [\fB\-finSv\fP] [\fB\-e\fP\~\fIextender\fP] [\fB\-I\fP\~\fIdir\fP] [\fB\-o\fP\~\fIprefix\fP] [\fB\-R\fP\~\fIremote\fP] [\fIfiles\fP]

.SH DESCRIPTION
\fIExplode\fP breaks a source file into separate compilation units.
The source text should contain control directives of the form
.br
	\fIdelimiter\fP\fIdirective\fP \fIargument\fP
.br
all of which must begin in column one of a source line.
.PP
A \fB\&MICE\fP \*(lqlibrary\*(rq or family is the least specific implementation of
an algorithm.  Unlike a run-time configured routine (like \fBqsort\fP(3))
a MICE family is configured via compile time macros: thus no
efficiency or functionality is lost due to a \*(lqbad fit\*(rq.
.PP
\fB\&MICE\fP families come packaged in two files: a large source file
and a small header file.
The C source file, for example, may be compiled with
a \-DTEST option to provide a simple test program.  The header file
is intended to provide all the function prototypes needed to use
the family.  Before the library can be configured for a new application
the user should run the C source through \fI\*(PN\fP.
Of course, any additional \fBmake\fP(1) descriptions, manual pages, and/or HTML
documents for the library may be kept as well.
.PP
The \fB\&Explode\fP directive opens a new output stream according to the
following \fIname\fP (eg. \*(lq/*@Explode \fIname\fP@*/\*(rq for C text).
The name is assumed to end at either the first white-space character or
the first occurrence of the last character in \fIdelimiter\fP.
The \fIname\fP is prefixed by \fIprefix\fP and extended by \fIextender\fP;
the resulting string is passed to \fBfopen\fP(3) as a file name.
.PP
The \fB\&Append\fP directive opens a new output stream according to the
following \fIname\fP.  The \fIname\fP is treated as if it were for
an \fB\&Explode\fP directive, but the file is opened for append and
no header is prefixed to it.
.PP
The \fB\&Header\fP directive instructs \fI\*(PN\fP to begin saving a
header that should be copied to the beginning of all the output files.
This directive has no argument (yet).  To remove this from the
output see \fB\-c\fP below.
.PP
The \fB\&Remove\fP directive diverts text to the null device until another
directive opens a new destination.
.PP
The \fB\&Message\fP directive diverts text to the invokers tty until
another directive opens a new destination.
.PP
The \fB\&Insert\fP directive copies its argument to the current output
stream.  The first non-white character of the argument is taken to be a
delimiter that marks the end of the string.
.PP
The \fB\&Shell\fP directive copies the output of the given \fBsh\fP(1)
command to the current output stream.
The shell command is enclosed in quotes (any non-white character).
.SH OPTIONS
The environment variable $\fB\&EXPLODE\fP is read for options before
the command-line option are processed.
.TP
.nf
\fB\-c\fP
.fi
Under this flag none of the exploded modules will have the Header
section included.  This is useful when they are to be subsequently
textually included in another file (via \fBcpp\fP(1) or the like).
.TP
.nf
\fB\-d\fP \fIdelimiter\fP
.fi
Use the given \fIdelimiter\fP rather than the default for the current
extender. Delimiters should
be short strings. The table below gives some of the default delimiters:
.br
.RS +6
.TS
c c c.
extension	language	default delimiter
c or h	C	/*@
k	K	/*@
f	FORTRAN	c@
p	Pascal	(*@
sh	sh	#@
csh	csh	#@
.TE
.RE
.TP
.nf
\fB\-e\fP \fIextender\fP
.fi
In an \fB\&Explode\fP directive, the \fIextender\fP will be
added to the \fBbasename\fP.  By default it is the extender
of the source file, or \*(lq.c\*(rq if no extender can be found.
.TP
.nf
\fB\-f\fP
.fi
Normally \fI\*(PN\fP will not overwrite existing files: if this option
is given it will.
.TP
.nf
\fB\-h\fP
.fi
Output a only usage summary.
.TP
.nf
\fB\-H\fP
.fi
Output only a short file format reminder.
.TP
.nf
\fB\-i\fP
.fi
Confirm any file over-write operations, like \fBrm\fP(1).
.TP
.nf
\fB\-I\fP \fIdir\fP
.fi
A colon separated directory path to search for source files.  The
default repository is displayed under \fB\-V\fP.
.TP
.nf
\fP\-n\fP
.fi
Print what would be done but don't build files.
This option implies the \fB\-v\fP option.
.TP
.nf
\fB\-o\fP \fIprefix\fP
.fi
In an \fB\&Explode\fP directive the \fIprefix\fP will be
placed before the \fIbasename\fP.  By default it is the basename of
the file being read.  a \fIprefix\fP of dash (\*(lq-\*(rq) directs
the output to \fIstdout\fP.
.TP
.nf
\fB\-R\fP \fIremote\fP[\fB:\fP\fIexplmux\fP]
.fi
Consult a \fBtcpmux\fP(8l) service on \fIremote\fP to find the aggregate file.
Operate on the stream locally.  The default \fIexplmux\fP is \*(lqexplmux\*(rq,
and it usually not changed by local site policy.  This option may
be disabled at compile time, if the local machine can't support
\fBgetaddrinfo\fP(3).  See \fBexplmux\fP(7l).
.TP
.nf
\fB\-s\fP
.fi
Output only the source module located (do not explode it into the
parts).
This option is often used to copy source modules to a temporary directory.
.TP
.nf
\fB\-S\fP
.fi
Just display the path to the source file we located.
.TP
.nf
\fB\-t\fP
.fi
Output only the table of contents from each source file, as \fBtar\fP(1)
does for any archive.
.TP
.nf
\fB\-T\fP
.fi
Only list the recognized file extenders.
.TP
.nf
\fB\-u\fP \fIunpack\fP
.fi
Cause \fI\*(PN\fP to unpack only the listed files. \fIUnpack\fP may
be a list of names separated by commas.  This option my be repeated
to concatenate the lists.
.TP
.nf
\fB\-v\fP
.fi
Output a running synopsis of \fI\*(PN\fP's actions.
.TP
.nf
\fB\-V\fP
.fi
Output only \fI\*(PN\fP's version information.
.SH EXAMPLES
.TP
.nf
\fI\*(PN\fP \-o /tmp/foo \-e .c kevin
.fi
To unpack the file \*(lqkevin\*(rq to /tmp/foo\fIbasename\fP.c
which should create /tmp/fooone.c and /tmp/footwo.c.
.TP
.nf
\fI\*(PN\fP \-o /tmp/avl \-u ins \-u del \-u init avl.c
.fi
Only unpack the insert, delete and init routines from the AVL source
file.
.TP
.nf
\fI\*(PN\fP \-t mk.c
.fi
Extract the catalog of available components from the mk.c module.
.TP
.nf
\fI\*(PN\fP \-o \- \-u utemp mk.c |${PAGER:\-more}
.fi
Output the unix domain socket code from the mktemp emulation library
to \fIstdout\fP, send that output to a pager.
.TP
.nf
\fI\*(PN\fP \-T |${PAGER:\-more}
.fi
Output the list of all the default markup expected for each extender.
.TP
.nf
\fI\*(PN\fP \-s avl.c
.fi
Copy the Adelson-Velskii Landis height-balanced tree module to
the current directory.
.TP
.nf
cp \-i `\fI\*(PN\fP \-sS avl.c` .
.fi
Same as above, but use \fBcp\fP(1) to avoid an accidental overwrite
of any local copy.
.TP
.nf
\fI\*(PN\fP \-s \-R explode.example.com avl.c
.fi
Download the same module from the local \fBexplmux\fP(7l) service.
.SH BUGS
There is no inverse of the \-c option: that is there is no (good) way
to just get the Header section of the archive.  Using \fBdiff\fP(1) to
get it is an unnecessary hack: you almost never need the header without
at least one function.
.SH AUTHOR
KS Braunsdorf, NonPlayer Character Guild
.br
mkcmd and no-spam on ksb.npcguild.org
.PP
This is a modification of the original Purdue code, by the original author.
Since Purdue doesn't distribute explode anymore -- you might have to live
with this version.
.SH "SEE ALSO"
.hlm 0
explode(5l), explmux(7l), make(1), cc(1), sh(1), mkcmd(1l),
basename(1)
