.\" $Id: explode.man,v 6.15 2013/10/22 19:33:12 ksb Exp $
.\" $Compile: Display%h
.\" $Display: ${groff:-groff} -Tascii -man %f |${PAGER:-less}
.\" $Display(*): ${groff:-groff} -T%s -man %f
.\" $Install: %b -mDeinstall %o %f && cp %f $DESTDIR${TOP:-/usr/local}/man/man5/explode.5
.\" $Deinstall: ${rm-rm} -f $DESTDIR${TOP:-/usr/local}/man/[cm]a[nt]5/explode.5*
.TH EXPLODE 5 LOCAL
.SH NAME
explode - file format for MICE archives
.SH DESCRIPTION
.ds PN "explode
.P
A \fB\&MICE\fP libraryor family is the most general form of
an algorithm. Unlike a run-time configured routine, like \fBqsort\fP(3),
a MICE family is configured via compile time macros: thus no
efficiency or functionality is lost due to a \*(lqbad fit\*(rq.
\fB\&MICE\fP families come packaged in two files:  a large source file
and a small header file.
.P
See \fBexplode\fP(1l) for a complete description \fI\*(PN\fP's usage.
.P
\fIExplode\fP directives are hidden from the native processor (viz.
the C compiler or mkcmd) in in-line comments.
An escape sequence including the native comment introducer and some
additional noise activates \fI\*(PN\fP as it scans the input file.
The delimiter is spelled backwards on the end of the line as a sanity check.
.P
Lexically explode looks at each source line as a potential command.
Only treating the input line as a command when it finds
the in-line comment sequence followed by a directive.
Directives may have a single parameter.
When this parameter is a string it is quoted by the first non-space
character after the directive keyword.
.P
Most commonly a C source file will only contain a few \fI\*(PN\fP
directives:
.RS
.nf
/*@Header@*/
#include <stdio.h>
#include "me.h"
#include "my_config.h"

/*@Insert \`extern int global_variable;\`@*/

/*@Explode routine1@*/
/*@Message "globals are in routine1"@*/
int global_variable;
int
routine1()
{
	...
}

/*@Explode routine2@*/
int
routine2()
{
}
.fi
.RE
.SH DIRECTIVES
.TP
.nf
/*@Message %This package requires the header file "tool_cfg.h".%@*/
.fi
Notify the user that the package requires a special header file.
.TP
.nf
/*@Header@*/
.fi
Divert the following lines to the common prologue for all
modules.
.TP
.nf
/*@Shell %echo "/* this was extracted on \`date\` by ${NAME-${USER-${LOGNAME-root}}} *""/"%@*/
.fi
Insert the date of extraction into the correct file.
This is normally done only once in each Header section.
.TP
.nf
/*@Insert %extern int errno;%@*/
.fi
Add a declaration for \fIerrno\fP to this module.
.TP
.nf
/*@Explode gen@*/
.fi
Copy the current Header to the \*(lqgen\*(rq module, begin diverting
text to that module.
.TP
.nf
/*@Remove@*/
.fi
Remove documentation or example usage from the file.  (Divert to /dev/null.)
.TP
.nf
/*@Append gen@*/
.fi
Redivert text to the \*(lqgen\*(rq module.
.TP
.nf
/*@Version: $Revision:... @*/
.fi
Provide a version header for \-t option.
.TP
.nf
c@Explode regress@c
.fi
A FORTRAN explode directive.
.SH CULTURE
.P
Most of the modules included in the \fB\&MICE\fP archive are coded
by a small group of NPC Guild pundits.
All these pundits follow a few naming and format rules that make
their code look similar in shape and style.
.SH "Variable names and Paints"
Most variable names start with a \fIpaint\fP which describes the
variables type (maybe not all the type, just the smell).
The rules for paints are described on the painted types Web page.
.SH "Layout of the code"
A MICE archive is broken into sections by convention and by \fI\*(PN\fP.
The first section has stuff in it that describes the archive itself,
its version, its \fBmk\fP marked lines, or some comments about the
usage and history of the part [citations for the algorithms used].
.P
That section ends with the \fI\*(PN\fP directive \fB\&Header\fP.
The header itself begins with a notice that the extracted part is
not to be edited because the real source is the whole part (which
is version controlled as a unit)!
.P
Each subsequent section is one of the other \fI\*(PN\fP hunks.
Each function that might stand on its own is broken into a separate
file.  Some functions are combined into a single file (for example
if they are mutually recursive).
.P
The last hunk in the file is usually a test driver.  Here is the
an example driver boiler plate:
.RS
.nf
/*@Remove@*/
#if defined(TEST)

/*@Explode main@*/

char *progname = "\fIpkgname\fP-main";

int
main(argc, argv)
{
	\fImain function\fP
}
/*@Remove@*/
#endif /* test driver */
.fi
.RE
.SH AUTHOR
Kevin Braunsdorf, NPC Guild.org
.br
mkcmd atspam-filter ksb dot npcguild.org
.P
This is a modification of the code the author wrote at Purdue.
Purdue has no (stated) interest in this software anymore.
.SH "SEE ALSO"
.hlm 0
explode(1l), mkcmd(1l), sh(1), mk(1l)
