# $Id: api_explmux.m,v 8.3 2013/06/10 16:32:57 ksb Exp $
# This API fetches files from explmux for explode and mkcmd		(ksb)

from '<stdio.h>'
from '<unistd.h>'
from '<sys/types.h>'
from '<sys/socket.h>'
from '<sys/uio.h>'
from '<netdb.h>'

%hi
#if !defined(HAVE_GETADDRINFO)
#define HAVE_GETADDRINFO	1
#endif

#if !defined(EXPLMUX_PORT)
#if EXPLMUX_SERVICES
#define EXPLMUX_PORT	"tcpmux"
#else
#define EXPLMUX_PORT	"1"		/* spec of local tcpmux service	*/
#endif
#endif

#if !defined(EXPLODE_MUX)
#define EXPLODE_MUX	"explmux"	/* name of service from above	*/
#endif

extern void VersionExplmux(FILE *fp);
#if HAVE_GETADDRINFO
extern FILE *Remote_Open(const char **ppcRemote, const char *pcFile, const char *pcMode, void *);
#define RemoteOpen(M1, M2, M3)	Remote_Open(M1, M2, M3, (void *)0)
#else
#define RemoteOpen(M1, M2, M3) (fprintf(stderr, "%s: getaddrinfo unavailable\n", progname), (FILE *)0)
#endif

%%
%i
/* Show the Customer (on stderr) the failed reply for each unavailable file?
 */
#if !defined(EXPLMUX_CARPS)
#define EXPLMUX_CARPS	0
#endif

const char acMuxService[] =
	EXPLODE_MUX;

%%
%c
/* mkcmd common version function					(ksb)
 */
void
VersionExplmux(FILE *fp)
{
#if HAVE_GETADDRINFO
	fprintf(fp, "%s: default tcmpux service: %s\n", progname, acMuxService);
#else
	fprintf(fp, "%s: missing getaddrinfo support for -R\n", progname);
#endif
}

#if HAVE_GETADDRINFO
/* replace \r or \r\n with \n in a C string				(ksb)
 */
static char *
NetToUnix(char *pcReply)
{
	register char *pc;

	if ((char *)0 == (pc = strchr(pcReply, '\r'))) {
		/* nada */
	} else if ('\n' == pc[1]) {
		NetToUnix(pc+2);
		memmove(pc, pc+1, strlen(pc+1)+1);
	} else {
		*pc = '\n';
	}
	return pcReply;
}

/* stop infinite redirection for RFC1078 + ksb's @ redirection
 */
struct ELMnode {
	const char *pcsaw;	/* hostname we tried			*/
	const char *pcport;	/* port we connected to			*/
	void *pvback;		/* really a (struct ELMnode *)		*/
};

/* consult a tcmpmux service at pcRemote				(ksb)
 * Remote may be in the form of host:service where 	service is the
 * name of the tcmpux service to consult.
 */
FILE *
Remote_Open(const char **ppcRemote, const char *pcFile, const char *pcMode, void *pvTrail)
{
	register const char *pcRemote, *pcColon;
	register int iL;
	auto const char *pcMuxSvc = acMuxService;
	auto struct addrinfo hints, *res, *res0;
	auto int iError, sFd;
	auto const char *pcCause = (char *)0;
	auto struct iovec aIOV[4];
	auto char acReply[MAXPATHLEN*2+sizeof("+\r\n")+1];
	auto char acEor[3] = "\r\n";
	static const char *pcPort = EXPLMUX_PORT;

	pcRemote = *ppcRemote;
	if ((char *)0 != (pcColon = strchr(pcRemote, ':')) && '\000' != pcColon[1]) {
		pcMuxSvc = pcColon+1;
	}
	if ((char *)0 != pcColon) {
		iL = pcColon - pcRemote;
		register char *pcMem;
		pcMem = strdup(pcRemote);
		pcMem[iL] = '\000';
		pcRemote = pcMem;
		*ppcRemote = pcRemote;
	}
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_NUMERICSERV;
	hints.ai_canonname = (char *)0;
	hints.ai_next = (struct addrinfo *)0;
	if (0 != (iError = getaddrinfo(pcRemote, pcPort, &hints, &res0))) {
		fprintf(stderr, "%s: getaddrinfo %s: %s\n", progname, pcRemote, gai_strerror(iError));
		exit(EX_NOHOST);
	}
	sFd = -1;
	for (res = res0; res; res = res->ai_next) {
		if (-1 == (sFd = socket(res->ai_family, res->ai_socktype, res->ai_protocol))) {
			pcCause = "socket";
			continue;
		}
		if (-1 != connect(sFd, res->ai_addr, res->ai_addrlen)) {
			break;
		}
		pcCause = "connect";
		close(sFd);
		sFd = -1;
	}
	freeaddrinfo(res0);
	if (-1 == sFd) {
		fprintf(stderr, "%s: tcpmux: %s: %s: %s\n", progname, pcCause, pcRemote, strerror(errno));
		exit(EX_NOINPUT);
	}
	aIOV[0].iov_base = (char *)pcMuxSvc;
	aIOV[0].iov_len = strlen(aIOV[0].iov_base);
	aIOV[1].iov_base = acEor;
	aIOV[1].iov_len = sizeof(acEor-1);
	if (aIOV[0].iov_len+aIOV[1].iov_len != writev(sFd, aIOV, 2)) {
		fprintf(stderr, "%s: tcpmux service write: %s\n", progname, strerror(errno));
		exit(EX_PROTOCOL);
	}
	/* read "+Go\r\n", "-Fail\r\n", or "@redir:port\r\n"
	 */
	if (-1 == (iL = read(sFd, acReply, sizeof(acReply)-1))) {
		fprintf(stderr, "%s: tcpmux closed: %s\n", progname, strerror(errno));
		exit(EX_PROTOCOL);
	}
	acReply[iL] = '\000';

	/* Redirected? Take that in stride, but check for redirection loops.
	 * @[host]:[port]
	 */
	if ('@' == acReply[0]) {
		register struct ELMnode *pELM;
		register char *pcColon;
		register void *pvScan;
		auto const char *pcDown;
		auto struct ELMnode ELMThis;

		ELMThis.pcsaw = pcRemote;
		ELMThis.pcport = pcPort;
		ELMThis.pvback = pvTrail;
		pvTrail = (void *)& ELMThis;

		if ((char *)0 != (pcColon = strchr(acReply, '\r'))) {
			*pcColon = '\000';
		}
		if ((char *)0 != (pcColon = strchr(acReply, '\n'))) {
			*pcColon = '\000';
		}
		if ((char *)0 == (pcColon = strchr(acReply, ':'))) {
			errno = ENOPROTOOPT;
			return (FILE *)0;
		}
		pcDown = acReply+1;
		if ('\000' == pcDown[0]) {
			pcDown = pcRemote;
		}
		*pcColon++ = '\000';
		if ('\000' != pcColon[0]) {
			pcPort = pcColon;
		}

		for (pvScan = pvTrail; (void *)0 != pvScan; pvScan = pELM->pvback) {
			pELM = (struct ELMnode *)pvScan;
			if (0 == strcmp(pELM->pcport, pcColon) && 0 == strcmp(pELM->pcsaw, pcDown)) {
				errno = ELOOP;
				return (FILE *)0;
			}
		}
		return Remote_Open(& pcDown, pcFile, pcMode, pvTrail);
	}
	if ('+' != acReply[0]) {
		acReply[iL] = '\000';
		NetToUnix(acReply);
#if EXPLMUX_CARPS
		fprintf(stderr, "%s: %s:%s: %s", progname, pcRemote, pcMuxSvc, acReply+1);
#endif
		errno = ENOENT;
		return (FILE *)0;
	}
	aIOV[0].iov_base = (char *)pcFile;
	aIOV[0].iov_len = strlen(aIOV[0].iov_base);
	aIOV[1].iov_base = acEor;
	aIOV[1].iov_len = sizeof(acEor-1);
	if (aIOV[0].iov_len+aIOV[1].iov_len != writev(sFd, aIOV, 2)) {
		fprintf(stderr, "%s: tcpmux filename write: %s\n", progname, strerror(errno));
		exit(EX_PROTOCOL);
	}
	if (-1 == (iL = read(sFd, acReply, sizeof(acReply)))) {
		fprintf(stderr, "%s: tcpmux service quit: %s\n", progname, strerror(errno));
		exit(EX_SOFTWARE);
	}
	if ('+' != acReply[0]) {
		acReply[iL] = '\000';
		NetToUnix(acReply);
#if EXPLMUX_CARPS
		fprintf(stderr, "%s: %s:%s: %s: %s", progname, pcRemote, pcMuxSvc, pcFile, acReply+1);
#endif
		errno = EACCES;
		return (FILE *)0;
	}
	return fdopen(sFd, pcMode);
}
#endif
